#include "operator.h"
#include "calculation.h"

#define USE_LAPACK
#ifdef USE_LAPACK
extern "C"
{
#include "cblas.h"
}
#endif

class Product : public Operator
{
public:
  Product ()
  { setDimensions (4);
    dimSizes [3] = 1; // woher soll das normal kommen ????
    operands . push_back (Operand (true, 0, 1));
    operands . push_back (Operand (true, 1, 2));
    operands . push_back (Operand (false, 0, 2, 3));
    splitDimensions = 3;
    label = "*";
  }

  void operate ()
  { Slab a = getOperandData (0), b = getOperandData (1), c = getOperandData (2);
#ifndef USE_LAPACK
//c . printFrame ();
    for (int i = 0;  i < a . getSize (0);  i ++)
      for (int j = 0;  j < b . getSize (1);  j ++)
      { Data r = 0;
        for (int k = 0;  k < a . getSize (1);  k ++)  r += a (i, k) * b (k, j);
        c (i, j) = r;
      }
#else
//a.printFrame();
//b.printFrame();
//c.printFrame();
//cout << endl;
    if (a . getSize (0) == 0 || b . getSize (1) == 0 || a . getSize (1) == 0)  return;
    bool aRowMajor = a . getStride (1) == 1;
    bool bRowMajor = b . getStride (1) == 1;
    bool cRowMajor = c . getStride (1) == 1;
    cblas_dgemm
      ( cRowMajor ? CblasRowMajor : CblasColMajor
      , aRowMajor != cRowMajor ? CblasTrans : CblasNoTrans
      , bRowMajor != cRowMajor ? CblasTrans : CblasNoTrans
      , a . getSize (0), b . getSize (1), a . getSize (1)
      , 1.0
      , a . getData (), aRowMajor ? a . getStride (0) : a . getStride (1)
      , b . getData (), bRowMajor ? b . getStride (0) : b . getStride (1)
      , 0.0
      , c . getData (), cRowMajor ? c . getStride (0) : c . getStride (1)
      );      
#endif
  }

  void calcDestSizes ()
  {
    dimSizes [3] = divisions [1];
    divisions [3] = divisions [1];
  }

protected:
  double calcCompCost ()
  { dimSizes [3] = divisions [1];
    divisions [3] = divisions [1];
    return COMP_COST * dividedSize (0) * dividedSize (1) * dividedSize (2);
  }
};


class GenerateMatrix : public Operator
{
public:
  GenerateMatrix (unsigned m, unsigned n)
  { setDimensions (2);
    operands . push_back (Operand (false, 0, 1));
    splitDimensions = 2;
    dimSizes [0] = m;  dimSizes [1] = n;
    label = "gen";
  }

  void operate ()
  { vector <int> pos;
    Slab a = getOperandData (0, & pos);
    for (int i = 0;  i < a . getSize (0);  i ++)
      for (int j = 0;  j < a . getSize (1);  j ++)
        a (i, j) = i + pos [0] + 2 * (j + pos [1]);
  }

protected:
  double calcCompCost ()  {return 0;}  
};


class ConsumeMatrix : public Operator
{
public:
  ConsumeMatrix ()
  { setDimensions (2);
    operands . push_back (Operand (true, 0, 1));
    splitDimensions = 2;
    label = "con";
  }

  void operate ()
  { Slab a = getOperandData (0);
    for (int i = 0;  i < a . getSize (0);  i ++)
    { for (int j = 0;  j < a . getSize (1);  j ++)
        cout << a (i, j) << ' ';
      cout << endl;
    }
  }

protected:
  double calcCompCost ()  {return 0;}
};


class PointwiseMult : public Pointwise
{
public:
  PointwiseMult (unsigned dim)
  : Pointwise (dim, ".")
    {}

  void operate ()
  { Slab a = getOperandData (0), b = getOperandData (1), c = getOperandData (2);
    for (int i = 0;  i < a . getSize (0);  i ++)
      for (int j = 0;  j < a . getSize (1);  j ++)
         c (i, j) = a (i, j) * b (i, j);
  }

protected:
  double calcCompCost ()
  { double c = 0;
    for (unsigned i = 0;  i < dimensions;  i ++)  c *= dividedSize (i);
    return COMP_COST * c;
  }
};


class PointwiseDiv : public Pointwise
{
public:
  PointwiseDiv (unsigned dim)
  : Pointwise (dim, "/")
    {}

  void operate ()
  { Slab a = getOperandData (0), b = getOperandData (1), c = getOperandData (2);
    for (int i = 0;  i < a . getSize (0);  i ++)
      for (int j = 0;  j < a . getSize (1);  j ++)
         c (i, j) = a (i, j) / (b (i, j) + 0.00001);
  }

protected:
  double calcCompCost ()
  { double c = 0;
    for (unsigned i = 0;  i < dimensions;  i ++)  c *= dividedSize (i);
    return COMP_COST * c;
  }
};


class AddFold3To2Operator : public Operator
{
public:
  AddFold3To2Operator ()
  { setDimensions (3);
    operands . push_back (Operand (true, 0, 1, 2));
    operands . push_back (Operand (false, 0, 1));
    splitDimensions = 2;
    label = "R";
  }

  void operate ()
  { Slab a = getOperandData (0), b = getOperandData (1);
    for (int i = 0;  i < a . getSize (0);  i ++)
      for (int j = 0;  j < a . getSize (1);  j ++)
      { Data e = a (i, j , 0);
        for (int k = 1;  k < a . getSize (2);  k ++)   e += a (i, j, k);
        b (i, j) = e;
      }
  }

protected:
  double calcCompCost ()
    {return COMP_COST * dividedSize (0) * dividedSize (1) * dividedSize (2);}
};


class AddReduce3To2Operator : public ReduceOperator
{
public:
  AddReduce3To2Operator ()
  { setDimensions (4);
    operands . push_back (Operand (true, 0, 1, 2));
    operands . push_back (Operand (false, 0, 3));
    splitDimensions = 3;
    label = "R";
    sourceDims = 3;
    destDims . push_back (0);
    destDims . push_back (3);
    destDims . push_back (3);
    reducedDims . push_back (false);
    reducedDims . push_back (false);
    reducedDims . push_back (true);
    //divisionLocked = true;
  }

  void operate ()
  { Slab a = getOperandData (0), b = getOperandData (1);
    for (int i = 0;  i < a . getSize (0);  i ++)
      for (int j = 0;  j < a . getSize (1);  j ++)
      { Data e = a (i, j , 0);
        for (int k = 1;  k < a . getSize (2);  k ++)   e += a (i, j, k);
        b (i, j) = e;
      }
  }

protected:
  double calcCompCost ()
  { return   COMM_COST * dimSizes [0] * dimSizes [1] * dimSizes [2]
           + COMP_COST * dividedSize (0) * dividedSize (1) * dividedSize (2);
  }
};


inline Variable & operator* (Variable & a, Variable & b)
{
  Calculation & cal = * a . calculation;
  Product * p = new Product;
  cal . addOperator (p);
  p -> plug (a, 0);
  p -> plug (b, 1);
  Variable * c = cal . newVariable (p, 2);
  AddFold3To2Operator * f = new AddFold3To2Operator;
  cal . addOperator (f);
  f -> plug (* c, 0);
  return * cal . newVariable (f, 1);
}


inline Variable & generateMatrix (unsigned m, unsigned n, Calculation & cal)
{
  GenerateMatrix * gm = new GenerateMatrix (m, n);
  cal . addOperator (gm);
  return * cal . newVariable (gm, 0);
}


inline void consumeMatrix (Variable & a)
{
  ConsumeMatrix * cm = new ConsumeMatrix;
  Calculation & cal = * a . calculation;
  cal . addOperator (cm);
  cm -> plug (a, 0);
}


inline Variable & pMult (Variable & a, Variable & b)
{
  Calculation & cal = * a . calculation;
  PointwiseMult * p = new PointwiseMult (2);
  cal . addOperator (p);
  p -> plug (a, 0);
  p -> plug (b, 1);
  return * cal . newVariable (p, 2);
}


inline Variable & pDiv (Variable & a, Variable & b)
{
  Calculation & cal = * a . calculation;
  PointwiseDiv * p = new PointwiseDiv (2);
  cal . addOperator (p);
  p -> plug (a, 0);
  p -> plug (b, 1);
  return * cal . newVariable (p, 2);
}



